#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.


from octobot_commons.databases.run_databases import run_databases_identifier
from octobot_commons.databases.run_databases import run_databases_provider
from octobot_commons.databases.run_databases import storage
from octobot_commons.databases.run_databases import abstract_run_databases_pruner
from octobot_commons.databases.run_databases import file_system_run_databases_pruner

from octobot_commons.databases.run_databases.run_databases_identifier import (
    RunDatabasesIdentifier,
)
from octobot_commons.databases.run_databases.run_databases_provider import (
    RunDatabasesProvider,
)
from octobot_commons.databases.run_databases.storage import (
    init_bot_storage,
    close_bot_storage,
)
from octobot_commons.databases.run_databases.abstract_run_databases_pruner import (
    AbstractRunDatabasesPruner,
)
from octobot_commons.databases.run_databases.file_system_run_databases_pruner import (
    FileSystemRunDatabasesPruner,
)
from octobot_commons.databases.run_databases.run_databases_pruning_factory import (
    run_databases_pruner_factory,
)


__all__ = [
    "RunDatabasesIdentifier",
    "RunDatabasesProvider",
    "init_bot_storage",
    "close_bot_storage",
    "AbstractRunDatabasesPruner",
    "FileSystemRunDatabasesPruner",
    "run_databases_pruner_factory",
]



# pylint: disable=W1203
#  Drakkar-Software OctoBot-Commons
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import logging
import hashlib
import cryptography.fernet as fernet

import octobot_commons
import octobot_commons.constants as commons_constants


def has_invalid_default_config_value(*config_values):
    """
    Check if config has invalid values
    :param config_values: the config values to check
    :return: the check result
    """
    return any(
        value in commons_constants.DEFAULT_CONFIG_VALUES for value in config_values
    )


def encrypt(data):
    """
    Basic encryption
    :param data: the data to encrypt
    :return: the encrypted data
    """
    try:
        return fernet.Fernet(octobot_commons.OCTOBOT_KEY).encrypt(data.encode())
    except Exception as global_exception:
        logging.getLogger().error(f"Failed to encrypt : {data}")
        raise global_exception


def decrypt(data, silent_on_invalid_token=False):
    """
    Basic decryption method
    :param data: the data to decrypt
    :param silent_on_invalid_token: if an error should be raised if a token is invalid
    :return: the decrypted data
    """
    try:
        return (
            fernet.Fernet(octobot_commons.OCTOBOT_KEY).decrypt(data.encode()).decode()
        )
    except fernet.InvalidToken as invalid_token_error:
        if not silent_on_invalid_token:
            logging.getLogger().error(
                f"Failed to decrypt : {data} ({invalid_token_error})"
            )
        raise invalid_token_error
    except Exception as global_exception:
        logging.getLogger().error(f"Failed to decrypt : {data} ({global_exception})")
        raise global_exception


def decrypt_element_if_possible(value_key, config_element, default="") -> str:
    """
    Return decrypted values, handles placeholder values
    :param value_key: the value key
    :param config_element: the config element
    :param default: the default value if no decrypt possible
    :return: True if the value can be decrypted
    """
    element = config_element.get(value_key, "")
    if element and not has_invalid_default_config_value(element):
        return decrypt(element)
    return default


def get_password_hash(password):
    """
    Returns the password's hex digest
    :param password: the password to hash
    :return: the hash digest
    """
    return hashlib.sha256(password.encode()).hexdigest()



# pylint: disable=R0913,R0914,C0415
#  Drakkar-Software OctoBot-Commons
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import os

import octobot_commons.databases.document_database_adaptors as adaptors
import octobot_commons.databases.implementations.cache_timestamp_database as cache_timestamp_database
import octobot_commons.databases.databases_util as databases_util
import octobot_commons.constants as common_constants
import octobot_commons.symbols.symbol_util as symbol_util
import octobot_commons.errors as common_errors
import octobot_commons.tree as tree


class CacheManager:
    """
    Manages cache as a global dict since caches can be accessed from live, backtesting and optimizers concurrently
    """

    CACHES = tree.BaseTree()
    DEFAULT_CONFIG_IDENTIFIER = "default"

    def __init__(self, database_adaptor=adaptors.TinyDBAdaptor):
        self.database_adaptor = database_adaptor

    def get_cache(
        self,
        tentacle,
        tentacle_name,
        exchange_name,
        symbol,
        time_frame,
        config_name,
        tentacles_setup_config,
        tentacles_requirements,
        cache_type=cache_timestamp_database.CacheTimestampDatabase,
        open_if_missing=True,
    ) -> tuple:
        """
        Returns the cache database associated to the given arguments. Creates/opens it if missing
        :param tentacle: tentacle to read configuration from to when creating/opening the associated database.
        Can be None when the database is open
        :param tentacle_name: name of the tentacle. Used to identify the database
        :param exchange_name: name of the exchange. Used to identify the database
        :param symbol: symbol. Used to identify the database
        :param time_frame: name of the timeframe. Used to identify the database
        :param config_name: name of the configuration (used in nested contexts). Used to identify the database
        :param tentacles_setup_config: the used tentacles_setup_config. Used to read configuration from to when
        creating/opening the associated database if the given tentacle instance doesn't already have its configuration
        :param tentacles_requirements: TentacleRequirementsTree associated to the given tentacle. Used to take
        nested calls into account when creating/opening databases
        :param cache_type: Type of the cache database. Used when creating/opening databases
        :param open_if_missing: When True, if the database is missing when asked for then it is created.
        :return: the cache database
        """
        identifier = config_name or self.DEFAULT_CONFIG_IDENTIFIER
        cache_path = [tentacle_name, exchange_name, symbol, time_frame, identifier]
        try:
            return self.__class__.CACHES.get_node(cache_path).node_value.get_database()
        except tree.NodeExistsError:
            if open_if_missing:
                if tentacle is None:
                    config_names = self.__class__.CACHES.get_children_keys(
                        cache_path[:-1]
                    )
                    available_config_names = (
                        f"Available configuration names: {config_names}. "
                        if config_names
                        else ""
                    )
                    raise common_errors.UninitializedCache(
                        f"No initialized cache for {tentacle_name} tentacle with config name: {config_name}. "
                        f"{available_config_names}"
                        f"The tentacle parameter must be set to get the associated cache database path"
                    )
                cache = self._open_or_create_cache_database(
                    tentacle,
                    exchange_name,
                    symbol,
                    time_frame,
                    tentacle_name,
                    identifier,
                    tentacles_setup_config,
                    cache_type,
                    tentacles_requirements,
                )
                self.__class__.CACHES.set_node_at_path(cache, None, cache_path)
                return cache.get_database()
            raise common_errors.NoCacheValue(
                f"Cache is initialized for {tentacle_name} on {exchange_name} "
                f"{symbol} {time_frame}"
            )

    def has_cache(
        self, tentacle_name, exchange_name, symbol, time_frame, config_name=None
    ):
        """
        Returns True if a cache database is open according to the given parameters
        :param tentacle_name: name of the tentacle
        :param exchange_name: name of the exchange
        :param symbol: associated symbol
        :param time_frame: name of the time frame
        :param config_name: name of the configuration
        """
        identifier = config_name or self.DEFAULT_CONFIG_IDENTIFIER
        try:
            return bool(
                self.__class__.CACHES.get_node(
                    [tentacle_name, exchange_name, symbol, time_frame, identifier]
                )
            )
        except tree.NodeExistsError:
            return False

    def get_cache_registered_requirements(
        self, tentacle_name, exchange_name, symbol, time_frame, config_name=None
    ):
        """
        Returns the TentacleRequirementsTree associated to the found cache database
        :param tentacle_name: name of the tentacle
        :param exchange_name: name of the exchange
        :param symbol: associated symbol
        :param time_frame: name of the time frame
        :param config_name: name of the configuration
        """
        identifier = config_name or self.DEFAULT_CONFIG_IDENTIFIER
        return self.__class__.CACHES.get_node(
            [tentacle_name, exchange_name, symbol, time_frame, identifier]
        ).node_value.tentacles_requirements

    def get_cache_previous_db_metadata(
        self, tentacle_name, exchange_name, symbol, time_frame, config_name=None
    ):
        """
        Returns the metadata associated of the previous cache database. Mostly used when including
        new dependencies to ensure metadata are preserved throughout databases
        :param tentacle_name: name of the tentacle
        :param exchange_name: name of the exchange
        :param symbol: associated symbol
        :param time_frame: name of the time frame
        :param config_name: name of the configuration
        """
        identifier = config_name or self.DEFAULT_CONFIG_IDENTIFIER
        return self.__class__.CACHES.get_node(
            [tentacle_name, exchange_name, symbol, time_frame, identifier]
        ).node_value.previous_db_metadata

    async def clear_cache(
        self,
        tentacle_name,
        exchange_name=common_constants.UNPROVIDED_CACHE_IDENTIFIER,
        symbol=common_constants.UNPROVIDED_CACHE_IDENTIFIER,
        time_frame=common_constants.UNPROVIDED_CACHE_IDENTIFIER,
        config_name=common_constants.UNPROVIDED_CACHE_IDENTIFIER,
    ):
        """
        Delete all the content of cache databases
        :param tentacle_name: name of the tentacle
        :param exchange_name: name of the exchange, all of them if left unspecified
        :param symbol: associated symbol, all of them if left unspecified
        :param time_frame: name of the time frame, all of them if left unspecified
        :param config_name: name of the configuration, all of them if left unspecified
        """
        try:
            for cache, _ in self._caches(
                tentacle_name, exchange_name, symbol, time_frame, config_name
            ):
                await cache.node_value.clear()
            return True
        except tree.NodeExistsError:
            return False

    async def reset_cache(
        self, tentacle_name, exchange_name, symbol, time_frame, config_name
    ):
        """
        Removes a cache database from cache manager. Closes but does not clear the database
        :param tentacle_name: name of the tentacle
        :param exchange_name: name of the exchange
        :param symbol: associated symbol
        :param time_frame: name of the time frame
        :param config_name: name of the configuration
        """
        identifier = config_name or self.DEFAULT_CONFIG_IDENTIFIER
        cache = self.__class__.CACHES.delete_node(
            [tentacle_name, exchange_name, symbol, time_frame, identifier]
        )
        await cache.node_value.close()

    async def close_cache(
        self,
        tentacle_name,
        exchange_name=common_constants.UNPROVIDED_CACHE_IDENTIFIER,
        symbol=common_constants.UNPROVIDED_CACHE_IDENTIFIER,
        time_frame=common_constants.UNPROVIDED_CACHE_IDENTIFIER,
        config_name=common_constants.UNPROVIDED_CACHE_IDENTIFIER,
        reset_cache_db_ids=False,
    ):
        """
        Closes but does not clear the associated databases. Removes cache databases from cache manager if
        reset_cache_db_ids
        :param tentacle_name: name of the tentacle
        :param exchange_name: name of the exchange, all of them if left unspecified
        :param symbol: associated symbol, all of them if left unspecified
        :param time_frame: name of the time frame, all of them if left unspecified
        :param config_name: name of the configuration, all of them if left unspecified
        :param reset_cache_db_ids: When True, removes the database from cache manager
        """
        try:
            to_remove_caches = []
            for cache, identifiers in self._caches(
                tentacle_name, exchange_name, symbol, time_frame, config_name
            ):
                await cache.node_value.close()
                to_remove_caches.append(identifiers)
            if reset_cache_db_ids:
                # remove cache from caches to force complete reopen of the cache db
                # (might be at a different place)
                for identifier in to_remove_caches:
                    self.__class__.CACHES.delete_node(identifier)
            return True
        except tree.NodeExistsError:
            return False

    async def reset(self):
        """
        Completely resets the cache manager closing and unregistering every cache
        """
        for cache, _ in self._caches():
            if cache.node_value.is_open():
                await cache.node_value.close()
        self.__class__.CACHES = tree.BaseTree()

    def _caches(
        self,
        tentacle_name=common_constants.UNPROVIDED_CACHE_IDENTIFIER,
        exchange_name=common_constants.UNPROVIDED_CACHE_IDENTIFIER,
        symbol=common_constants.UNPROVIDED_CACHE_IDENTIFIER,
        time_frame=common_constants.UNPROVIDED_CACHE_IDENTIFIER,
        config_name=common_constants.UNPROVIDED_CACHE_IDENTIFIER,
    ):
        path = []
        for element in (tentacle_name, exchange_name, symbol, time_frame, config_name):
            if element == common_constants.UNPROVIDED_CACHE_IDENTIFIER:
                break
            path.append(element)
        if self.__class__.CACHES.get_children_keys(path):
            for (
                cache,
                identifiers,
            ) in self.__class__.CACHES.get_nested_children_with_path(path):
                # avoid caches that have no node value (might be remains of already cleared caches)
                if cache.node_value is not None:
                    yield cache, identifiers
        # no cache value
        return []

    def _open_or_create_cache_database(
        self,
        tentacle,
        exchange,
        symbol,
        time_frame,
        tentacle_name,
        identifier,
        tentacles_setup_config,
        cache_type,
        tentacles_requirements,
    ):
        cache_full_path = self.get_cache_or_build_path(
            tentacle,
            exchange,
            symbol,
            time_frame,
            tentacle_name,
            identifier,
            tentacles_setup_config,
            tentacles_requirements,
        )
        cache_dir = os.path.split(cache_full_path)[0]
        if not os.path.exists(cache_dir):
            os.makedirs(cache_dir)
        return self._open_cache_database(
            cache_full_path, cache_type, tentacles_requirements
        )

    def _open_cache_database(self, file_path, cache_type, tentacles_requirements):
        """
        Override to use another cache database or adaptor
        :return: the cache database class
        """
        return databases_util.CacheWrapper(
            file_path, cache_type, self.database_adaptor, tentacles_requirements
        )

    def get_cache_or_build_path(
        self,
        tentacle,
        exchange_name,
        symbol,
        time_frame,
        tentacle_name,
        config_name,
        tentacles_setup_config,
        tentacles_requirements,
    ):
        """
        Returns the cache path associated to the given arguments. Use local cache when available otherwise
        recompute the whole path (time consuming)
        :param tentacle: tentacle use to build the path. Required if path is not in cache already
        :param tentacle_name: name of the tentacle
        :param exchange_name: name of the exchange
        :param symbol: associated symbol, all of them if left unspecified
        :param time_frame: name of the time frame, all of them if left unspecified
        :param config_name: name of the configuration, all of them if left unspecified
        :param tentacles_setup_config: the used tentacles_setup_config. Used to read configuration from to when
        creating/opening the associated database if the given tentacle instance doesn't already have its configuration
        :param tentacles_requirements: TentacleRequirementsTree associated to the given tentacle. Used to take
        nested calls into account when creating/opening databases
        """
        identifier = config_name or self.DEFAULT_CONFIG_IDENTIFIER
        try:
            return self.__class__.CACHES.get_node(
                [tentacle_name, exchange_name, symbol, time_frame, identifier]
            ).node_value.get_path()
        except tree.NodeExistsError:
            sanitized_pair = symbol_util.merge_symbol(symbol) if symbol else symbol
            required_tentacles = tentacles_requirements.get_all_required_tentacles(
                False
            )
            # ensure tentacles requirements are snapshotting the configuration that was used to build the
            # cache identifier
            tentacles_requirements.synchronize_tentacles_config()
            identifying_tentacles = [tentacle] + required_tentacles
            # warning: very slow, should be called as rarely as possible
            code_hash, config_hash = self._tentacles_hashes(
                identifying_tentacles, tentacles_setup_config
            )
            return os.path.join(
                common_constants.USER_FOLDER,
                common_constants.CACHE_FOLDER,
                tentacle_name,
                exchange_name,
                sanitized_pair,
                time_frame,
                code_hash,
                config_hash,
                common_constants.CACHE_FILE,
            )

    @staticmethod
    def _tentacles_hashes(identifying_tentacles, tentacles_setup_config) -> (str, str):
        try:
            import octobot_tentacles_manager.api

            return (
                octobot_tentacles_manager.api.get_code_hash(identifying_tentacles)[
                    : common_constants.CACHE_HASH_SIZE
                ],
                octobot_tentacles_manager.api.get_config_hash(
                    identifying_tentacles, tentacles_setup_config
                )[: common_constants.CACHE_HASH_SIZE],
            )
        except ImportError as err:
            raise ImportError(
                "octobot_tentacles_manager is required to use cache"
            ) from err



# cython: language_level=3
#  Drakkar-Software OctoBot-Commons
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.


cdef class Symbol:
    cdef public str symbol_str
    cdef public str base
    cdef public str quote
    cdef public str settlement_asset
    cdef public str identifier
    cdef public str strike_price
    cdef public str option_type
    cdef public str full_symbol_regex
    cdef public str market_separator
    cdef public str settlement_separator

    cpdef object parse_symbol(self, str symbol_str)
    cpdef tuple base_and_quote(self)
    cpdef bint is_linear(self)
    cpdef bint is_inverse(self)
    cpdef str merged_str_symbol(self, str market_separator=*, str settlement_separator=*)



# pylint: disable=C0301, R0904
#  Drakkar-Software OctoBot-Commons
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import os

try:
    import tinydb
    import tinydb.storages
    import tinydb.middlewares
    import tinydb.table
except ImportError:
    pass

import octobot_commons.logging as commons_logging
import octobot_commons.constants as constants
import octobot_commons.errors as errors
import octobot_commons.databases.document_database_adaptors.abstract_document_database_adaptor as abstract_document_database_adaptor


class TinyDBAdaptor(abstract_document_database_adaptor.AbstractDocumentDatabaseAdaptor):
    """
    TinyDBAdaptor is an AbstractDatabaseAdaptor implemented using tinydb: a minimal python only
    local document database.
    Warning: loads the whole file in RAM and must be closed to ensure writing
    """

    DEFAULT_WRITE_CACHE_SIZE = 5000

    def __init__(self, file_path: str, cache_size: int = None, **kwargs):
        """
        TinyDBAdaptor constructor.
        :param file_path: path to the database file
        :param cache_size: size of the in memory cache (number of operations before updating the file
        :param kwargs: unused
        """
        super().__init__(file_path)
        self.database = None
        self.cache_size = cache_size

    def initialize(self):
        """
        Initialize the database: opens the database file.
        """
        middleware = tinydb.middlewares.CachingMiddleware(tinydb.storages.JSONStorage)
        middleware.WRITE_CACHE_SIZE = self.cache_size or self.DEFAULT_WRITE_CACHE_SIZE
        try:
            self.database = tinydb.TinyDB(self.db_path, storage=middleware)
        except FileNotFoundError as err:
            raise errors.DatabaseNotFoundError(
                f'Can\'t open database at "{self.db_path}"'
            ) from err

    @staticmethod
    def is_file_system_based() -> bool:
        """
        Returns True when this database is identified as a file in the current file system,
        False when it's managed by a database server
        """
        return True

    @staticmethod
    def get_db_file_ext() -> str:
        """
        Returns the database file extension. Implemented in file system based databases
        """
        return constants.TINYDB_EXT

    @staticmethod
    async def create_identifier(identifier):
        """
        Initialize the identifier by creating it in the database
        """
        if not os.path.exists(identifier):
            os.makedirs(identifier)

    @staticmethod
    async def identifier_exists(identifier, is_full_identifier) -> bool:
        """
        Returns True when the given identifier is part of an existing database identifier
        :param identifier: the identifier to look into
        :param is_full_identifier: when True, only check identifiers that don't have sub identifiers.
        When False, only check identifiers that have sub identifiers
        """
        return (
            os.path.isfile(identifier)
            if is_full_identifier
            else os.path.isdir(identifier)
        )

    @staticmethod
    async def get_sub_identifiers(identifier, ignored_identifiers):
        """
        Returns an iterable over the existing sub-identifiers under the given identifier
        """
        for folder in os.scandir(identifier):
            if (
                await TinyDBAdaptor.identifier_exists(folder, False)
                and folder.name not in ignored_identifiers
            ):
                yield folder.name

    @staticmethod
    async def get_single_sub_identifier(identifier, ignored_identifiers) -> str:
        """
        Returns the name of the only sub-identifier at a given parent identifier, None otherwise
        example use: get the name of the only exchange the backtesting happened on if it only ran on a single exchange,
        """
        exchange_folders = [
            folder.name
            for folder in os.scandir(identifier)
            if os.path.isdir(folder) and folder.name not in ignored_identifiers
        ]
        return exchange_folders[0] if len(exchange_folders) == 1 else None

    def get_uuid(self, document) -> int:
        """
        Returns the uuid of the document
        :param document: the document
        """
        return document.doc_id

    async def select(self, table_name: str, query, uuid=None) -> list:
        """
        Select data from the table_name table
        :param table_name: name of the table
        :param query: select query
        :param uuid: id of the document
        """
        if uuid is None:
            return (
                self.database.table(table_name).search(query)
                if query
                else self.database.table(table_name).all()
            )
        return self.database.table(table_name).get(doc_id=uuid)

    async def tables(self) -> list:
        """
        Select tables
        """
        return list(self.database.tables())

    async def insert(self, table_name: str, row: dict) -> int:
        """
        Insert dict data into the table_name table
        :param table_name: name of the table
        :param row: data to insert
        """
        return self.database.table(table_name).insert(row)

    async def upsert(self, table_name: str, row: dict, query, uuid=None) -> int:
        """
        Insert or update dict data into the table_name table
        :param table_name: name of the table
        :param row: data to insert
        :param query: select query
        :param uuid: id of the document
        """
        if uuid is None:
            return self.database.table(table_name).upsert(row, query)
        return self.database.table(table_name).upsert(
            tinydb.table.Document(row, doc_id=uuid)
        )

    async def insert_many(self, table_name: str, rows: list) -> list:
        """
        Insert multiple dict data into the table_name table
        :param table_name: name of the table
        :param rows: data to insert
        """
        return self.database.table(table_name).insert_multiple(rows)

    async def update(self, table_name: str, row: dict, query, uuid=None) -> list:
        """
        Select data from the table_name table
        :param table_name: name of the table
        :param row: data to update
        :param query: select query
        :param uuid: id of the document
        """
        if uuid is None:
            return self.database.table(table_name).update(row, query)
        return self.database.table(table_name).update(
            tinydb.table.Document(row, doc_id=uuid)
        )

    async def update_many(self, table_name: str, update_values: list) -> list:
        """
        Update multiple values from the table_name table
        :param table_name: name of the table
        :param update_values: values to update
        """
        return self.database.table(table_name).update_multiple(update_values)

    async def delete(self, table_name: str, query, uuid=None) -> list:
        """
        Delete data from the table_name table
        :param table_name: name of the table
        :param query: select query
        :param uuid: id of the document
        """
        if uuid is None:
            if query is None:
                return self.database.drop_table(table_name)
            return self.database.table(table_name).remove(query)
        return self.database.table(table_name).remove(doc_ids=(uuid,))

    async def count(self, table_name: str, query) -> int:
        """
        Counts documents in the table_name table
        :param table_name: name of the table
        :param query: select query
        """
        return self.database.table(table_name).count(query)

    async def query_factory(self):
        """
        Creates a new empty select query
        """
        return tinydb.Query()

    async def hard_reset(self):
        """
        Completely reset the database
        """
        await self.close()
        os.remove(self.db_path)
        self.initialize()

    async def flush(self):
        """
        Flushes the database cache
        """
        return self.database.storage.flush()

    async def close(self):
        """
        Closes the database
        """
        try:
            return self.database.close()
        except AttributeError:
            # when self.database didn't open properly
            pass
        except TypeError as err:
            commons_logging.get_logger(str(self)).exception(
                err,
                True,
                f"Error when writing database, this is probably due to a script that "
                f"is saving a non json-serializable value: {err}",
            )



#  Drakkar-Software OctoBot-Commons
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import mock
import pytest

import octobot_commons.logging as logging
import octobot_commons.logging.logging_util as logging_util


@pytest.fixture
def logger():
    return logging.get_logger("test")


@pytest.fixture
def call_wrapper():
    callback_mock = mock.Mock()

    class Wrapper:
        def __init__(self):
            self.callback_mock = callback_mock

        def other_callback(self, *args, **kwargs):
            callback_mock(*args, **kwargs)
    return Wrapper()


def test_register_error_callback():
    def other_call_back():
        pass
    logging.BotLogger.register_error_callback(logging_util._default_callback)
    assert logging_util._ERROR_CALLBACK is logging_util._default_callback
    logging.BotLogger.register_error_callback(other_call_back)
    assert logging_util._ERROR_CALLBACK is other_call_back


def test_error(logger, call_wrapper):
    logging.BotLogger.register_error_callback(call_wrapper.other_callback)
    logger.error("err")
    call_wrapper.callback_mock.assert_called_once_with(None, "err")
    call_wrapper.callback_mock.reset_mock()

    logger.error("err", skip_post_callback=True)
    call_wrapper.callback_mock.assert_not_called()


def test_exception(logger, call_wrapper):
    logging.BotLogger.register_error_callback(call_wrapper.other_callback)
    err = None
    def raiser():
        def other():
            1/0
        other()
    try:
        raiser()
    except Exception as exc:
        err = exc
        logger.exception(err, True, "error")
    call_wrapper.callback_mock.assert_called_once_with(err, "error")
    call_wrapper.callback_mock.reset_mock()

    logger.exception(err, True, "error", skip_post_callback=True)
    call_wrapper.callback_mock.assert_not_called()
